VAGRANT_TEMPLATE = """# -*- mode: ruby -*-
# vi: set ft=ruby :

VAGRANTFILE_API_VERSION = "2"

Vagrant.configure(VAGRANTFILE_API_VERSION) do |config|

  config.vm.box = "{box}"

end"""


DOCKER_TEMPLATE = """FROM {image}

USER root

# install packages
ENV DEBIAN_FRONTEND=noninteractive
RUN apt-get update
RUN apt-get install -y {packages}

# init user and ssh
EXPOSE 22
RUN useradd --create-home --shell /bin/bash -g sudo {user}
RUN chown -R {user}:sudo /home/{user}
RUN chmod u+s /usr/bin/sudo
RUN echo "{user} ALL=(ALL) NOPASSWD:ALL" > /etc/sudoers.d/{user} && chmod 0440 /etc/sudoers.d/{user}
USER {user}

WORKDIR /home/{user}
COPY {keyfile} .ssh/authorized_keys

USER root
COPY {keyfile} /root/.ssh/authorized_keys
RUN mkdir -p /run/sshd && \\
  chmod 755 /run/sshd

ENTRYPOINT []

CMD /usr/sbin/sshd; \\
  while true; do sleep 1m; done
"""


# TODO: proper template generation for alpine
DOCKER_ALPINE_TEMPLATE = """FROM {image}

USER root

# install packages
RUN apk update
RUN apk add --no-cache python3
RUN apk add --no-cache musl-dbg
# install gdb
RUN apk add --no-cache gdb
# install ssh server support and keys
RUN apk add --no-cache openssh
# install sudo
RUN apk add --no-cache sudo

EXPOSE 22
RUN chmod u+s /usr/bin/sudo
RUN adduser -h /home/{user} -s /bin/ash -g sudo -D {user}
RUN echo "{user} ALL=(ALL) NOPASSWD:ALL" > /etc/sudoers.d/{user} && chmod 0440 /etc/sudoers.d/{user}
RUN echo "{user}:{user}" | chpasswd

USER {user}

WORKDIR /home/{user}

COPY {keyfile} .ssh/authorized_keys

USER root
COPY {keyfile} /root/.ssh/authorized_keys
RUN ssh-keygen -A
RUN mkdir -p /run/sshd && \
  chmod 755 /run/sshd

ENTRYPOINT []

CMD /usr/sbin/sshd; \
  while true; do sleep 1m; done

"""



import os
import sys

import pwnlib.log

from vagd.virts.pwngd import Pwngd

_GENERATE_KEYPAIR = 'ssh-keygen -q -t ed25519 -f {keyfile} -N ""'
log = pwnlib.log.getLogger(None)
log.setLevel(pwnlib.log.logging.INFO)


def _ensure_log():
  """
  ensure the correct logger is set
  """
  global log
  if "pwn" in sys.modules:
    from pwn import log as pwnlog

    log = pwnlog


def info(info: str):
  """
  log info with pwntools
  :param info: info to log
  """
  _ensure_log()
  log.info(info)


def debug(debug: str):
  """
  log debug with pwntools
  :param debug: debug to log
  """
  _ensure_log()
  log.debug(debug)


def warn(warn: str):
  """
  log warn with pwntools
  :param warn: warn to log
  """
  _ensure_log()
  log.warn(warn)


def error(error: str):
  """
  log error with pwntools
  :param error: error to log
  """
  _ensure_log()
  log.error(error)


def progress(progress: str) -> pwnlib.log.Progress:
  """
  log progress with pwntools
  :param progress: progress to log
  """
  _ensure_log()
  return log.progress(progress)


def generate_keypair():
  """
  generate a keypair in .vagd directory
  """
  if not (
    os.path.exists(Pwngd.KEYFILE) and os.path.exists(Pwngd.KEYFILE + ".pub")
  ):
    info("No Keypair was found. Generating new keypair")
    os.system(_GENERATE_KEYPAIR.format(keyfile=Pwngd.KEYFILE))


def is_port_in_use(port: int) -> bool:
  """
  check if a port is currently used
  :param port: port to check
  :return: if the port is already used
  """
  import socket

  with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    return s.connect_ex(("localhost", port)) == 0


def first_free_port(start: int = 2222, tries: int = 101) -> int:
  """
  returns first free port starting from start (max increments = tries)
  :param start: start of port search
  :param tries: number of tries to increment ports
  :return: first free port
  """
  for i in range(tries):
    port = start + i
    if not is_port_in_use(port):
      return port

  error(f"No free port inside range {start}-{start + tries}")



import os
from typing import Dict, List

import docker

from vagd import helper, templates
from vagd.box import Box
from vagd.virts.pwngd import Pwngd
from vagd.virts.shgd import Shgd


class Dogd(Shgd):
  """
  | Docker virtualization for pwntools

  :param binary: binary to execute
  :param image: docker base image
  :param user: name of user on docker container
  :param forward: Dictionary of forwarded ports, needs to follow docker api format: 'hostport/(tcp|udp)' : guestport
  :param packages: packages to install on the container
  :param symbols: additionally install libc6 debug symbols (also updates libc6)
  :param ex: if experimental features, e.g. alpine, gdbserver should be enabled
  :param rm: remove container after exit
  :param alpine: if the conainter is alpine (also autochecks image name)
  :param fast: mounts libs locally for faster symbol extraction (experimental) NOT COMPATIBLE WITH ALPINE
  :param kwargs: parameters to pass through to super

  | SSH from cmd

  .. code-block:: bash

      vagd ssh
      # or
      ssh -o "StrictHostKeyChecking=no" -i ~/.share/local/vagd/keyfile -p $(cut .vagd/docker.lock -d":" -f 2) vagd@0.0.0.0

  | connect with docker exec

  .. code-block:: bash

     docker exec -it $(cut ./.vagd/docker.lock -d":" -f 1) /bin/bash

  | Kill from cmd:

  .. code-block:: bash

      vagd clean
      #or
      docker kill $(cut ./.vagd/docker.lock -d":" -f 1)

  | Docker containers are automatically removed after they stop
  | Docker images need to be manually removed from docker
  | Dockerfiles are stored in home directory to allow caching ~/.share/local/vagd/docker/<image>/Dockerfile

  .. code-block:: bash

      docker images # list images
      docker rmi <id> # remove correct image
  """

  _image: str
  _name: str
  _user: str
  _port: int
  _packages: List[str]
  _client: docker.client
  _id: str
  _dockerdir: str
  _dockerfile: str
  _isalpine: bool
  _gdbsrvport: int
  _rm: bool
  _ex: bool
  _forward: Dict[str, int]
  _symbols: bool

  VAGD_PREFIX = "vagd-"
  TYPE = "dogd"
  DOCKERHOME = Pwngd.HOME_DIR + "docker/"
  DEFAULT_USER = "vagd"
  DEFAULT_PORT = 2222
  DEFAULT_IMAGE = Box.DOCKER_NOBLE

  DEFAULT_PACKAGES = Pwngd.DEFAULT_PACKAGES + ["openssh-server"]
  LOCKFILE = Pwngd.LOCAL_DIR + "docker.lock"

  def __init__(
    self,
    binary: str,
    image: str = DEFAULT_IMAGE,
    user: str = DEFAULT_USER,
    forward: Dict[str, int] = None,
    packages: List[str] = None,
    symbols=True,
    rm=True,
    ex: bool = False,
    fast: bool = False,
    alpine: bool = False,
    **kwargs,
  ):
    self._image = image
    self._name = Dogd.VAGD_PREFIX + os.path.basename(binary)
    self._packages = list(Dogd.DEFAULT_PACKAGES)

    if symbols:
      helper.warn(
        f"installing {Pwngd.LIBC6_DEBUG} might update libc binary, consider using symbols=False"
      )
      self._packages.append(Pwngd.LIBC6_DEBUG)

    self._isalpine = alpine or "alpine" in image

    if packages is not None:
      if self._isalpine:
        helper.error("additional package installation not supported for alpine")
    else:
      # trigger package detection in Pwngdb
      packages = list()

    self._gdbsrvport = -1
    self._dockerdir = Dogd.DOCKERHOME + f"{self._image}/"
    if not (
      os.path.exists(Dogd.DOCKERHOME) and os.path.exists(self._dockerdir)
    ):
      os.makedirs(self._dockerdir)
    self._dockerfile = self._dockerdir + "Dockerfile"
    self._user = user
    self._forward = forward
    self._rm = rm
    self._ex = ex
    self._symbols = symbols
    if self._isalpine and not self._ex:
      helper.error("Docker alpine images requires experimental features")
    if self._forward is None:
      self._forward = dict()

    self._vm_setup()

    super().__init__(
      binary=binary,
      user=self._user,
      port=self._port,
      packages=packages,
      ex=ex,
      fast=fast,
      symbols=False,
      gdbsrvport=self._gdbsrvport,
      **kwargs,
    )

  def _create_dockerfile(self):
    helper.info(f"create new Dockerfile at {self._dockerfile}")
    if not os.path.exists(Pwngd.KEYFILE):
      helper.generate_keypair()

    if not os.path.exists(self._dockerdir + "keyfile.pub"):
      os.link(Pwngd.PUBKEYFILE, self._dockerdir + "keyfile.pub")
    template = (
      templates.DOCKER_ALPINE_TEMPLATE
      if self._isalpine
      else templates.DOCKER_TEMPLATE
    )

    with open(self._dockerfile, "w") as dockerfile:
      dockerfile.write(
        template.format(
          image=self._image,
          packages=" ".join(self._packages),
          user=self._user if self._user != "root" else Dogd.DEFAULT_USER,
          keyfile=os.path.basename(self._dockerdir + "keyfile.pub"),
        )
      )

  def _create_docker_instance(self):
    self.is_new = True
    helper.info("starting docker instance")
    self._port = helper.first_free_port(Dogd.DEFAULT_PORT)
    self._forward.update({"22/tcp": self._port})
    if self._isalpine:
      self._gdbsrvport = helper.first_free_port(Pwngd.STATIC_GDBSRV_PORT)
      self._forward.update({f"{self._gdbsrvport}/tcp": self._gdbsrvport})

    dir = os.path.dirname(os.path.realpath(__file__))
    with open(dir[: dir.rfind("/")] + "/res/seccomp.json", "r") as seccomp_file:
      seccomp_rules = seccomp_file.read().strip()

    container = self._client.containers.run(
      self._bimage,
      name=self._name,
      ports=self._forward,
      detach=True,
      remove=self._rm,
      security_opt=[f"seccomp:{seccomp_rules}"],
    )
    self._id = container.id
    helper.info(f"started docker instance {container.short_id}")
    with open(Dogd.LOCKFILE, "w") as lockfile:
      lockfile.write(
        f"{container.id}:{str(self._port)}:{str(self._gdbsrvport)}"
      )

  def _build_image(self):
    build_progress = helper.progress("building docker image")
    hash = self._image.find("@")
    if hash != -1:
      tag = self._image[:hash].replace(":", "_")
      # add first 8 characters of hash
      tag += self._image[self._image.rfind(":") :][:8]
    else:
      tag = self._image

    if self._symbols:
      if ":" not in tag:
        tag += ":"
      else:
        tag += "_"
      tag += "symbols"

    bimage = self._client.images.build(
      path=os.path.dirname(self._dockerfile), tag=f"vagd/{tag}"
    )[0]

    build_progress.success("done")

    return bimage

  def _vm_create(self):
    self._lock(Dogd.TYPE)

    if not os.path.exists(Pwngd.LOCAL_DIR):
      os.makedirs(Pwngd.LOCAL_DIR)

    # enfore changes to Dockerfile are always rebuild by docker
    self._create_dockerfile()

    self._bimage = self._build_image()

    self._create_docker_instance()

  def _vm_setup(self) -> None:
    self._client = docker.from_env()
    if not os.path.exists(Dogd.LOCKFILE):
      helper.info(
        f"No Lockfile {Dogd.LOCKFILE} found, creating new Docker Instance"
      )
      self._vm_create()
    else:
      with open(Dogd.LOCKFILE, "r") as lockfile:
        data = lockfile.readline().split(":")
        self._id = data[0]
        self._port = int(data[1])
        if self._isalpine:
          self._gdbsrvport = int(data[2])
      if not self._client.containers.list(filters={"id": self._id}):
        helper.info(
          f"Lockfile {Dogd.LOCKFILE} found, container not running, creating new one"
        )
        self._vm_create()
      else:
        helper.info(
          f"Lockfile {Dogd.LOCKFILE} found, Docker Instance f{self._client.containers.get(self._id).short_id}"
        )



class Empty:
  def __getattr__(self, name):
    """
    create empty function wrapper

    :param name: name of attribute
    :rtype: None
    """
    return lambda *args, **kwargs: None


class GDB:
  # Weird type hint that allows type hinting from stub files
  # stubs from: https://github.com/python/typeshed/tree/main/stubs/gdb/gdb
  # include subdir from other remote: https://stackoverflow.com/a/30386041
  def __new__(cls, target):
    # type:() -> gdb.GDB
    """
    returns GDB out of target if it has attribute gdb

    :param target: target to check for wrapper
    :return: returns Empty wrapper or gdb python api
    :rtype: gdb.GDB
    """
    if hasattr(target, "gdb"):
      return target.gdb
    else:
      return Empty()



import importlib.metadata
import os
import stat
import sys
from typing import Dict, List, Optional

import typer
from rich.console import Console
from rich.syntax import Syntax


# prevents term.init
from vagd.virts.dogd import Dogd
from vagd.virts.pwngd import Pwngd
from vagd.virts.qegd import Qegd
from vagd.virts.vagd import Vagd

DOGD_BOX = "Box.DOCKER_UBUNTU"
DOGD = "vm = Dogd(BINARY, image={box}, {args})  # Docker"
QEGD_BOX = "Box.QEMU_UBUNTU"
QEGD = "vm = Qegd(BINARY, img={box}, {args})  # Qemu"
SHGD = "vm = Shgd(BINARY, user='user', host='localhost', port=22, {args})  # SSH"

# deprecated
VAGD_BOX = "Box.VAGRANT_JAMMY64"
VAGD = "vm = Vagd(BINARY, {box}, {args})  # Vagrant"

AD_ENV = """# ad envs
IS_AD  = os.getenv('TARGET_IP') is not None           # running on ad
IP     = os.getenv('TARGET_IP', IP)                   # remote ip
EXTRA  = json.loads(os.getenv('TARGET_EXTRA', '[]'))  # flag ids"""

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})

err_console = Console(stderr=True)
console = Console()

quote = lambda x: f"'{x}'"


def _version(value: bool) -> None:
  if value:
    version = importlib.metadata.version("vagd")
    typer.echo(f"VAGD v{version}")
    raise typer.Exit()


@app.callback()
def main(
  version: Optional[bool] = typer.Option(
    None,
    "--version",
    "-v",
    help="Show current vagd version and exit.",
    callback=_version,
    is_eager=True,
  ),
) -> None:
  pass


def add_virt(
  dependencies: List[str],
  vms: List[str],
  dependency: str,
  template: str,
  args: Dict[str, str],
  multi=False,
  box="",
):
  dependencies.append(dependency)
  args_str = ", ".join(f"{k}={v}" for k, v in args.items())
  vm = template.format(box=box, args=args_str)
  vms.append(("# " if multi else "") + vm)


def _info(binary, color=True) -> str:
  from pwnlib.elf.elf import ELF
  from pwnlib.term import text

  red = text.red if color else str
  green = text.green if color else str

  exe = ELF(binary, checksec=False)
  out = list()
  out.append("Arch:".ljust(12))
  arch = green("-".join((exe.arch, str(exe.bits), exe.endian)))
  out.append(arch)
  out.append("\n")
  out.append(exe.checksec(color=color))
  out.append("\n")
  out.append("Comment:".ljust(12))
  if exe.get_section_by_name(".comment") is not None:
    comment = green(exe.section(".comment").replace(b"\0", b"").decode())
  else:
    comment = red("No Comment")

  out.append(comment)
  return "".join(out)


@app.command()
def template(
  binary: Optional[str] = typer.Argument("", help="Binary to Exploit"),
  ip: Optional[str] = typer.Argument("", help="Ip or Domain of the remote target"),
  port: Optional[int] = typer.Argument(0, help="port of the remote target"),
  output_exploit: Optional[bool] = typer.Option(
    False, "-e", help="output file of the template (also add +x) to exploit.py"
  ),
  output: Optional[str] = typer.Option(
    "", "-o", help="output file of the template (also add +x), default stdout"
  ),
  libc: Optional[str] = typer.Option("", "--libc", "-l", help="add libc to template"),
  libs: Optional[bool] = typer.Option(False, "--libs", help="download libraries from virt"),
  files: Optional[List[str]] = typer.Option([], "--files", "-f", help="add files to remote"),
  aslr: Optional[bool] = typer.Option(
    False, "--aslr", "-a", help="enable gdb ASLR (default: disabled for gdb)"
  ),
  dogd: Optional[bool] = typer.Option(
    False, "--dogd", "--docker", "-d", help="create docker template"
  ),
  image: Optional[str] = typer.Option(DOGD_BOX, "--image", help="docker image to use"),
  qegd: Optional[bool] = typer.Option(False, "--qegd", "--qemu", "-q", help="create qemu template"),
  img: Optional[str] = typer.Option(QEGD_BOX, "--img", help="qemu cloud image to use"),
  vagd: Optional[bool] = typer.Option(
    False, "--vagd", "--vagrant", help="DEPRECATED: create vagrant template"
  ),
  vbox: Optional[str] = typer.Option(VAGD_BOX, "--vbox", help="vagrant box to use"),
  shgd: Optional[bool] = typer.Option(False, "--shgd", "--ssh", "-s", help="create ssh template"),
  local: Optional[bool] = typer.Option(False, "--local", help="create local template"),
  ad: Optional[bool] = typer.Option(False, "--ad", help="create an ad compatible template"),
  root: Optional[bool] = typer.Option(False, "--root", "-r", help="create a root environment"),
  no_aliases: Optional[bool] = typer.Option(
    False, "--no-aliases", help="no aliases in the template"
  ),
  no_info: Optional[bool] = typer.Option(False, "--no-info", help="no binary info"),
):
  """
  creates a template
  """
  if image != DOGD_BOX:
    dogd = True
    image = quote(image)

  if img != QEGD_BOX:
    qegd = True
    img = quote(img)

  if vbox != VAGD_BOX:
    vagd = True
    vbox = quote(vbox)

  templatePath = os.path.dirname(os.path.realpath(__file__))
  templateChunks = list()
  aliasesPath = templatePath + "/res/aliases.txt"
  templatePath += "/res/template.txt"
  multi = False

  if not any((dogd, qegd, vagd, shgd)):
    dogd = qegd = True

  if sum((dogd, qegd, vagd, shgd)) > 1:
    multi = True

  env = {}

  if libc:
    files.append(libc)

  dependencies = []
  vms = []
  args = dict()

  if root:
    args["user"] = "'root'"

  if libs:
    args["libs"] = True
    if not libc:
      libc = "./libs/libc.so.6"

  if files:
    args["files"] = "[" + ",".join(f"'{file}'" for file in files) + "]"

  modules = list()
  if ad:
    modules.append("json")

  args["ex"] = "True"
  args["fast"] = "True"

  if dogd:
    add_virt(dependencies, vms, "Dogd", DOGD, args, box=image)
  if qegd:
    add_virt(dependencies, vms, "Qegd", QEGD, args, multi, box=img)
  if vagd:
    add_virt(dependencies, vms, "Vagd", VAGD, args, multi, box=vbox)
  if shgd:
    add_virt(dependencies, vms, "Shgd", SHGD, args, multi)

  with open(aliasesPath, "r") as aliases_file:
    aliases = aliases_file.read()

  with open(templatePath, "r") as templateFile:
    for line in templateFile.readlines():
      if libc and line.startswith("# libc"):
        templateChunks.append(line[2:])
      else:
        templateChunks.append(line)

  if not no_info:
    try:
      info = "# " + "\n# ".join(_info(binary, color=False).splitlines()) + "\n"
    except Exception:
      info = ""
  else:
    info = ""

  template = "".join(templateChunks).format(
    "{}",
    cmd_args=" ".join(sys.argv[1:]),
    modules="\n".join(f"import {module}" for module in modules),
    dependencies=", ".join(dependencies),
    aliases=aliases if not no_aliases else "",
    binary=quote(binary),
    ip=quote(ip),
    port=str(port),
    env=repr(env),
    ad_env=AD_ENV if ad else "",
    vms=("\n" + " " * 4).join(vms),
    libc=quote(libc),
    aslr=repr(aslr),
    is_local=True if local else "args.LOCAL",
    is_ad=" or IS_AD" if ad else "",
    info=info,
  )

  if output_exploit:
    output = "exploit.py"
  if output:
    with open(output, "w") as exploitFile:
      exploitFile.write(template)
    current_permissions = os.stat(output).st_mode
    new_permissions = current_permissions | (stat.S_IXUSR)
    os.chmod(output, new_permissions)
  else:
    syntax = Syntax(template, "python", theme="ansi_dark")
    console.print(syntax)


@app.command()
def info(
  binary: str = typer.Argument(..., help="Binary to analyse"),
):
  """
  analyses the binary, prints checksec and .comment (often includes Distro and Compiler info)
  """

  print(_info(binary))


def _get_type() -> str:
  if os.path.exists(Pwngd.LOCKFILE):
    with open(Pwngd.LOCKFILE) as lfile:
      return lfile.read()
  sys.stderr.write("no vagd instance is running\n")
  exit(1)


def _exec(cmd: str, env: Dict = None):
  if env is None:
    env = os.environ
  else:
    env.update(os.environ)
  os.execvpe("sh", ("sh", "-c", cmd), env)


def _ssh(port, user):
  os.system(f'ssh -o "StrictHostKeyChecking=no" -i {Pwngd.KEYFILE} -p {port} {user}@0.0.0.0')


@app.command()
def ssh(
  user: Optional[str] = typer.Option(None, "--user", "-u", help="ssh user"),
):
  """
  ssh to current vagd instance (must be in exploit dir)
  """
  typ = _get_type()
  if typ == Dogd.TYPE:
    if user is None:
      user = Dogd.DEFAULT_USER
    with open(Dogd.LOCKFILE, "r") as lfile:
      port = lfile.read().split(":")[1]
      _ssh(int(port), user)
  elif typ == Qegd.TYPE:
    if user is None:
      user = Qegd.DEFAULT_USER
    with open(Qegd.LOCKFILE, "r") as lfile:
      _ssh(int(lfile.read()), user)
  elif typ == Vagd.TYPE:
    os.system(f"VAGRANT_CWD={Pwngd.LOCAL_DIR} vagrant ssh")
  else:
    sys.stderr.write(f"Unknown type in {Pwngd.LOCKFILE}: {typ}\n")
    exit(1)


def _scp(
  port: int,
  user: str,
  source: str,
  target: str,
  recursive: bool,
  keyfile: str = Pwngd.KEYFILE,
):
  options = "-R" if recursive else ""
  if ":" in source:
    file = source[source.find(":") + 1 :]
    source = f"{user}@0.0.0.0:{file}"
  if ":" in target:
    file = target[target.find(":") + 1 :]
    target = f"{user}@0.0.0.0:{file}"
  os.system(f"scp -P {port} -o StrictHostKeyChecking=no -i {keyfile} {options} {source} {target}")


@app.command()
def scp(
  source: str = typer.Argument(..., help="source file"),
  target: str = typer.Argument("vagd:./", help="target file"),
  recursive: bool = typer.Option(False, "-r", "--recursive", help="recursive copy"),
  user: Optional[str] = typer.Option(None, "--user", "-u", help="ssh user"),
):
  """
  scp to from current vagd instance (must be in exploit dir)
  """
  typ = _get_type()
  if typ == Dogd.TYPE:
    if user is None:
      user = Dogd.DEFAULT_USER
    with open(Dogd.LOCKFILE, "r") as lfile:
      port = lfile.read().split(":")[1]
      _scp(int(port), user, source, target, recursive)
  elif typ == Qegd.TYPE:
    if user is None:
      user = Qegd.DEFAULT_USER
    with open(Qegd.LOCKFILE, "r") as lfile:
      _scp(int(lfile.read()), user, source, target, recursive)
  elif typ == Vagd.TYPE:
    import vagrant

    v = vagrant.Vagrant(os.path.dirname(Vagd.VAGRANTFILE_PATH))
    _scp(2222, v.user(), source, target, recursive, v.keyfile())
  else:
    sys.stderr.write(f"Unknown type in {Pwngd.LOCKFILE}: {typ}\n")
    exit(1)


@app.command()
def clean():
  """
  clean current vagd instance (stop/kill/remove/destroy)
  """
  typ = _get_type()
  if typ == Dogd.TYPE:
    if os.path.exists(Dogd.LOCKFILE):
      with open(Dogd.LOCKFILE, "r") as lockfile:
        data = lockfile.readline().split(":")
        id = data[0]
      import docker

      client = docker.from_env()
      if not client.containers.list(filters={"id": id}):
        sys.stderr.write(f"Lockfile {Dogd.LOCKFILE} found, container not running\n")
        exit(1)
      else:
        container = client.containers.get(id)
        typer.echo(f"Lockfile {Dogd.LOCKFILE} found, Docker Instance f{container.short_id}")
        container.kill()
  elif typ == Qegd.TYPE:
    os.system("kill $(pgrep qemu)")
  elif typ == Vagd.TYPE:
    import vagrant

    v = vagrant.Vagrant(os.path.dirname(Vagd.VAGRANTFILE_PATH))
    v.halt()
    v.destroy()
    os.remove(Vagd.VAGRANTFILE_PATH)
  else:
    sys.stderr.write(f"Unknown type in {Pwngd.LOCKFILE}: {typ}\n")
    exit(1)
  os.remove(Pwngd.LOCKFILE)


def start():
  app()



class Box:
  """constants class for recommended images and boxes"""

  QEMU_NOBLE = "https://cloud-images.ubuntu.com/noble/current/noble-server-cloudimg-amd64.img"
  QEMU_JAMMY = "https://cloud-images.ubuntu.com/jammy/current/jammy-server-cloudimg-amd64.img"
  QEMU_FOCAL = "https://cloud-images.ubuntu.com/focal/current/focal-server-cloudimg-amd64.img"
  QEMU_BIONIC = "https://cloud-images.ubuntu.com/bionic/current/bionic-server-cloudimg-amd64.img"
  QEMU_UBUNTU = QEMU_NOBLE

  QEMU_NOBLE_ARM = "https://cloud-images.ubuntu.com/noble/current/noble-server-cloudimg-arm64.img"
  QEMU_JAMMY_ARM = "https://cloud-images.ubuntu.com/jammy/current/jammy-server-cloudimg-arm64.img"

  DOCKER_NOBLE = "ubuntu:noble"
  DOCKER_JAMMY = "ubuntu:jammy"
  DOCKER_FOCAL = "ubuntu:focal"
  DOCKER_BIONIC = "ubuntu:bionic"
  DOCKER_XENIAL = "ubuntu:xenial"
  DOCKER_UBUNTU = DOCKER_NOBLE

  DOCKER_I386_FOCAL = "i386/ubuntu:focal"
  DOCKER_I386_BIONIC = "i386/ubuntu:bionic"
  DOCKER_I386_XENIAL = "i386/ubuntu:xenial"

  DOCKER_ALPINE_320 = "alpine:3.20"
  DOCKER_ALPINE = DOCKER_ALPINE_320

  # DEPRECATED
  VAGRANT_JAMMY64 = "ubuntu/jammy64"
  VAGRANT_FOCAL64 = "ubuntu/focal64"
  VAGRANT_BIONIC64 = "ubuntu/bionic64"
  VAGRANT_XENIAL64 = "ubuntu/xenial64"
